#include <stdlib.h>
#include <stdio.h>
#include <sys/signal.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#define NO_VLD_INDEX -1
#define SIGDEFT SIGTERM

#define LIST  1
#define	NOT_DEF  0
#define	SIGNAL  -1

struct signal {
	const char *name;
	int num;
};
	
struct signal sig_list[] = {
	{"0", 0},
	{"HUP", SIGHUP},
	{"INT", SIGINT},
	{"QUIT", SIGQUIT},
	{"ABRT", SIGABRT},
	{"KILL", SIGKILL},
	{"ALRM", SIGALRM},
	{"TERM", SIGTERM},
	{ NULL, 0 }
};
/**/
int 
get_idx_in_sig_list_by_name(const char *signame)
{
	struct signal *ptr;
	
	if (signame != NULL &&
	(signame[0] == 'S' || signame[0] == 's') &&
	(signame[1] == 'I' || signame[1] == 'i') &&
	(signame[2] == 'G' || signame[2] == 'g'))
		signame += 3;
		
	if (signame != NULL) 	
		for (ptr = sig_list; ptr->name != NULL; ptr++) 
			if (!strcasecmp(signame, ptr->name))
				return (ptr - sig_list);
	return NO_VLD_INDEX;
	
}

int 
get_idx_in_sig_list_by_num(int signum)
{
	struct signal *ptr;		
	
	for (ptr = sig_list; ptr->name != NULL; ptr++) 
		if (ptr->num == signum)
			return (ptr - sig_list);
			
	return NO_VLD_INDEX;	
}

#define GET_INDEX(x,s) ((errno) ? (get_idx_in_sig_list_by_name((s))): (get_idx_in_sig_list_by_num((x))))
#define GET_NUM_SIG(inx) (sig_list[inx].num)
#define GET_NAME_SIG(inx) (sig_list[inx].name)

/**/
void 
print_all()
{
	struct signal *ptr;
	
	printf("%d)%s\n", sig_list->num, sig_list->name);
	
	for (ptr = sig_list + 1 ; ptr->name != NULL; ptr++)
		printf("%d)SIG%s\n", ptr->num, ptr->name);	
}

void
print(int is_signame, int index)
{
	if (is_signame)
		printf("%s\n", GET_NAME_SIG(index));
	else 
		printf("%d\n", GET_NUM_SIG(index));	
}

int 
matoi(char *str)
{
	int res = 0, sign = 1;
	errno = 0;
		
	for (; (str != NULL) && ((*str) != 0) && (isblank(*str)); str++);
		
	if ((str != NULL) && ((*str) == '-')) {
		str++;
		sign = -1;
	}
	if ((str == NULL) || ((*str) == 0))
		goto ERROR;
	for (; (str != NULL) && ((*str) != 0); str++) {
		if (!isdigit(*str))
			goto ERROR;
		res *= 10;
		res += (int)((*str) - '0');
	}		

	return sign*res;
ERROR:
	errno = 1;
	return 0;

}

void 
mkill(pid_t pid, int index, char *progname)
{
	int signum = GET_NUM_SIG(index);
	if (kill(pid, signum) == -1)	
		switch (errno) {
			case EINVAL:
				fprintf(stderr, "%s:(%d) - invalid signal specification\n",
					progname, signum);
				break;
			case ESRCH:
				fprintf(stderr, "%s:(%d) - No such process\n", progname, pid);
				break;
			case EPERM:
				fprintf(stderr, "%s:(%d) - Operation not permitted\n", progname, pid);
				break;
		}
}


int 
main(int argc, char **argv)
{
	int i = 1;
	char *c = NULL;
	int value;
	pid_t curpid = 0;
	int option = NOT_DEF;
	int index = get_idx_in_sig_list_by_num(SIGDEFT);
	if (argc < 2)
		goto ERROR;
	else {
	
		while (i < argc) {
			
			if (option == LIST) {
				value = matoi(argv[i]); 			
				index = GET_INDEX(value, argv[i]);
				if (index != NO_VLD_INDEX)
					print(errno ? 0:1, index);
				else 
					fprintf(stderr,"%s:(%s) - invalid signal specification\n",
						argv[0], argv[i]);
				i++;
			} else if (option == SIGNAL) {

				curpid = matoi(argv[i]);
				
				if (!errno)								
					mkill(curpid, index, argv[0]);
				else
					fprintf(stderr,"%s:(%s) - arguments must be process or job IDs\n",
						argv[0], argv[i]);
				i++;
			} else 	if (option == NOT_DEF) {		
				c = argv[i];
				if ((c[0] == '-') && (c[1] != '\0')) {
					c++;
			 		if ((c[0] == '-') && (c[1] == '\0')) {
				 		option = SIGNAL;
						i++;
 					} else if ((c[0] == 'l') && (c[1] == '\0')) {
 						if (i+1 == argc) 
							print_all();
						else
						 	option = LIST;				 		
						i++;	
 					} else if ((c[0] == 's') && (c[1] == '\0')) {
 						if (i+1 == argc) {
							fprintf(stderr,"%s:-s:requires an argument\n",
								argv[0]);
							break;
						}
						value = matoi(argv[i+1]);
						index = GET_INDEX(value, argv[i+1]);
						if (index == NO_VLD_INDEX) {
							fprintf(stderr,
								"%s:(%s) - invalid signal specification\n",
								argv[0], argv[i+1]);
							break;
						}					
						if (i+2 == argc) 
							goto ERROR;
						option = SIGNAL;
						i+=2;
 					} else {	
				 		value = matoi(c);
				 		index = GET_INDEX(value, c);
				 		if (index == NO_VLD_INDEX){
							fprintf(stderr,
								"%s:(%s) - invalid signal specification\n",
								argv[0], c);
							break;
						}
						if (i+1 == argc) 
							goto ERROR;
						option = SIGNAL;					
						i++;
				 	}
 				} else if ((c[0] == '-') && (c[1] == '\0')) {
 					fprintf(stderr,
						"%s:(%s) - invalid signal specification\n",
						argv[0], c);
					break;
 				} else 
				 	option = SIGNAL;
			}
		}
	}
	return 0;
ERROR:
	printf("%s: [-s signum or signame] | [-signum ] | [-signame] pid\n\t"
		"or -l [signum or signame]\n",argv[0]);
	return 1;
}

